/*
 * this file is included from doc2ipf, doc2html and doc2info
 *
 * MUST include after termdoc.c (since it redefines fgets() )
 *
 * it contains functions needed to handle xrefs, most of them from
 *     doc2rtf (most likely) by Maurice Castro
 *  or doc2ipf by Roger Fearick
 *  or doc2html by Russel Lang
 *
 * I have modified the functions a little to make them more flexible 
 * (lookup returns list instead of list->line) or let them work with all
 * four programs (adding parameters three parameters to refs).
 *
 * I switched the search order of lookup. Makes more sense to me
 *
 * Stefan Bodewig 1/29/1996
 */

struct LIST
{
	int level;
	int line;
	char *string;
	struct LIST *next;
        struct LIST *prev;
	};

struct LIST *list = NULL;
struct LIST *head = NULL;

struct LIST *keylist = NULL;
struct LIST *keyhead = NULL;

int maxlevel = 0;       /* how deep are the topics nested? */
int listitems = 0;      /* number of topics */

#if 0
/* none of these prototypes are required, since there are no
 * forward calls made. Normally prototypes are fine, but because
 * xmalloc defn depends on circumstances, compile failed on
 * a non-ansi compiler I was using. [yes, I ought to come up with
 * a better fix !]
 */
void parse __P((FILE *a));
struct LIST *lookup __P((char*));
struct LIST *lkup_by_number __P((int line));
void list_free __P(());
void refs __P((int l, FILE *f, char *start, char *end, char *format));
void *xmalloc __P((size_t size));
#endif

/* for debugging (invoke from gdb !) */

void dump_list()
{
	struct LIST *element = head;
	while (element)
	{
		fprintf(stderr, "%p level %d, line %d, \"%s\"\n", element, element->level, element->line, element->string);
		element = element->next;
	}
}


#ifdef ANSI_C
void *xmalloc(size_t size)
{
	void *p=malloc(size);
#else
char *xmalloc(size)             /* think this is correct for K&R C */
unsigned int size;
{
	char *p=malloc(size);
#endif
	if (p)
		return p;
	fprintf(stderr, "Malloc failed\n");
	exit(1);
}
	
/* scan the file and build a list of line numbers where particular levels are */
void parse(a)
FILE *a;
{
    static char line[MAX_LINE_LEN];
    char *c;
    int lineno=0;
    int lastline=0;

    while (fgets(line,MAX_LINE_LEN,a)) 
    {
      lineno++;
      if (isdigit(line[0]))
      {
        listitems++;
        if (list == NULL) {   
            head = (list = (struct LIST *) xmalloc(sizeof(struct LIST)));
            list->prev = NULL;
        } else {
            list->next = (struct LIST *) xmalloc(sizeof(struct LIST));
            list->next->prev = list;
            list = list->next;
            list->next = NULL;
        }
        list->line = lastline = lineno;
        list->level = line[0] - '0';
        list->string = (char *) xmalloc (strlen(line)+1);
        c = strtok(&(line[1]),"\n");
        strcpy(list->string, c);
        list->next = NULL;
        if (list->level > maxlevel)
            maxlevel = list->level;
      }
      if (line[0]=='?')
      {
        if (keylist == NULL) {   
            keyhead = (keylist = (struct LIST *) xmalloc(sizeof(struct LIST)));
            keylist->prev = NULL;
        } else {
            keylist->next = (struct LIST *) xmalloc(sizeof(struct LIST));
            keylist->next->prev = keylist;
            keylist = keylist->next;
        }
        keylist->line = lastline;
        keylist->level = list->level;
        c = strtok(&(line[1]),"\n");
        if( c == NULL || *c == '\0' ) c = list->string ;
        keylist->string = (char *) malloc (strlen(c)+1);
        strcpy(keylist->string, c);
        keylist->next = NULL;
      }
    }
    rewind(a);
    }

/* look up a iopic in text reference */
/*
 * Original version from doc2rtf (|| ipf || html) scanned keylist before list.
 * This way we get a reference to `plot` for the topic `splot` instead
 * of one to `splot`. Switched them -SB.
 */
struct LIST *
lookup(s)
char *s;
{
    char *c;
    char tokstr[MAX_LINE_LEN];
    char *match; 
    int l;

    strcpy(tokstr, s);

    /* first try titles */
    match = strtok(tokstr, " \n\t");
    if( match == NULL ) {
        fprintf( stderr, "Erros in lookup()\n" ) ;
        fprintf( stderr, "Possible missing link character (`) near above line number\n" ) ;
        exit(3) ;
        }
    l = 0; /* level */
    
    list = head;
    while (list != NULL)
    {
        c = list->string;
        while (isspace(*c)) c++;
        if (!strcmp(match, c)) 
        {
            l = list->level;
            match = strtok(NULL, "\n\t ");
            if (match == NULL)
            {
                return(list);
                }
            }
        if (l > list->level)
            break;
        list = list->next;
        }

    /* then try the ? keyword entries */
    keylist = keyhead;
    while (keylist != NULL)
    {
        c = keylist->string;
        while (isspace(*c)) c++;
        if (!strcmp(s, c)) return(keylist);
        keylist = keylist->next;
        }

    return(NULL);
    }

/*
 * find title-entry for keyword-entry
 */
struct LIST *lkup_by_number(line)
    int line; 
{
    struct LIST *run = head;
    
    while (run->next && run->next->line <= line)
        run = run->next;
        
    if (run->next)
        return run;
    else
        return NULL;
}

/*
 * free the whole list (I never trust the OS -SB)
 */
void list_free __P((void)) {
    struct LIST *run;
    
    for(run = head; run->next; run = run->next)
        ;
    for(run = run->prev; run; run = run->prev) {
        free(run->next->string);
        free(run->next);
    }
    free(head->string);
    free(head);

    for(run = keyhead; run->next; run = run->next)
        ;
    for(run = run->prev; run; run = run->prev) {
        free(run->next->string);
        free(run->next);
    }
    free(keyhead->string);
    free(keyhead);
}


/* search through the list to find any references */
/* 
 * writes a menu of all subtopics of the topic located at l
 * format must contain %s for the title of the subtopic and may contain
 * a %d for the line number of the subtopic (used by doc2html and doc2rtf
 */
void
refs(l, f, start, end, format)
int l;
FILE *f;
char *start, *end, *format;
{
    int curlevel, i;
    char *c;
    int inlist = FALSE;

    /* find current line */
    list = head;
    while (list->line != l)
		list = list->next;
    curlevel = list->level;
    list = list->next;        /* look at next element before going on */
    if (start != NULL && list != NULL)
    {
	inlist = TRUE;
	fprintf(f, "%s", start);
	}

    while (list != NULL)
    {
        /* we are onto the next topic so stop */
        if (list->level == curlevel)
            break;
        /* these are the next topics down the list */
        if (list->level == curlevel+1)
        {
            c = list->string;
	    while (isspace(*c)) c++;
	    if (format != NULL) {
                for (i=0; format[i] != '%' && format[i] != '\0'; i++)
                    ;
                if (format[i] != '\0')
                    if (format[i+1] == 'd')
    	                fprintf(f, format, list->line, c);
    	            else {
    	                ++i;
                        for (; format[i] != '%' && format[i] != '\0'; i++)
                            ;
                        if (format[i] != '\0')
    	                    fprintf(f, format, c, list->line);
    	                else
                            fprintf(f, format, c);
                        }
                }
            }
        list = list->next;
        }
	if (inlist && end)
	    fprintf(f,"%s", end);
    }



